/*****************************************************************************
 * game.cpp - General game loop class
 *
 *****************************************************************************/
#include "SDL.h"
#include "game.h"
#include "freecnc.h" // for the game definitions
#include "vqa.h"
#include "common.h"
#include "loadingscreen.h"
#include "actioneventqueue.h"
#include "map.h"
#include "sidebar.h"
#include "input.h"
#include "cursor.h"
#include "logger.h"
#include "playerpool.h"
#include "config.h"
#include "netconnection.h"
#include "memtrack.h"

int Game::mapscaleq = -1;
/** Constructor, loads the map, sidebar and such. plays briefing and actionmovie
 * @param the graphicsengine.
 * @param the soundengine.
 * @param the name of the mission to play.
 */
Game::Game(GraphicsEngine *gfxengine, SoundEngine *sndengine)
{
    ConfigType config;
    VQAMovie *mov;
    char* message,*tmp;
    INIFile* fileini;
    LoadingScreen *loadscreen;
    config = getConfig();
    /* set the pointer to the gfx engine */
    this->gfxengine = gfxengine;
    this->sndengine = sndengine;
    try {
        fileini = new INIFile("files.ini");
    } catch(int) {
        return;
    }
    tmp = new char[8];
    sprintf(tmp,"play%i",config.gamenum);
    message = fileini->readString("general",tmp,"TD");
    delete[] tmp;
    if (!sndengine->createPlayList(message)) {
        logger->error("Could not create playlist!\n");
        throw new GameError();
    }
    delete[] message;
    delete fileini;
    loadscreen = new LoadingScreen(gfxengine);
    gamemode = config.gamemode;
    connection = NULL;
    if (gamemode == 2) {
        tmp = new char[64];
        sprintf(tmp,"Connecting to server: %s",config.serveraddr);
        loadscreen->setCurrentTask(tmp);
        delete[] tmp;

        try {
            connection = new NetConnection(config.serveraddr, config.serverport);
        } catch(int) {
            delete loadscreen;
            throw new GameError();
        }
        // after connection sending login data
        loadscreen->setCurrentTask("Sending Login Data");
        connection->login(VERSION, config.nick, config.mside, config.side_colour);
    }
    /* reset the tickcounter, should be a function in the class conatining the
     * counter */
    loadscreen->setCurrentTask("Creating the ActionEventQueue");
    evQueue = new ActionEventQueue();
    /* load the map */
    loadscreen->setCurrentTask("Loading the map (this might take a while)");
    try {
        ccmap = new CnCMap(evQueue, connection, loadscreen);
    } catch (LoadMapError* e) {
        delete e;
        delete loadscreen;
        // loadmap will have printed the error
        throw new GameError();
    }
    delete loadscreen;
    ccmap->getUSPool()->setsfxeng(sndengine);
    if (gamemode == 0) {
        /* play briefing */
        try {
            mov = new VQAMovie(gfxengine, sndengine, ccmap->getMissionData().brief);
            mov->play();
            delete mov;
        } catch (VQAError *vqaerr) {
            delete vqaerr;
        }
        try {
            mov = new VQAMovie(gfxengine, sndengine, ccmap->getMissionData().action);
            mov->play();
            delete mov;
        } catch (VQAError *vqaerr) {
            delete vqaerr;

        }
    }
    /* init sidebar */
    try {
        sidebar = new Sidebar(evQueue, ccmap->getPlayerPool()->getLPlayer(),
                              gfxengine->getHeight(), ccmap->getMissionData().theater);
    } catch (SidebarError* e) {
        delete e;
        throw new GameError();
    }
    /* init cursor */
    cursor = new Cursor();
    /* init the input functions */
    input = new Input(sidebar, ccmap, gfxengine->getWidth(),
                      gfxengine->getHeight(), gfxengine->getMapArea(),
                      cursor, sndengine);
    //sprintf(message, "Loaded map \"%s\"", missionname);
    //gfxengine->postMessage(message);
}

/** Destructor, frees up some memory */
Game::~Game()
{
    /*char *terminate = NULL;
    int len;*/
    if (gamemode == 2) {
        logger->note("Disconnecting from server...");
        connection->logout();
        delete connection;
        logger->note("done\n");
    }
    delete evQueue;
    delete input;
    delete sidebar;
    delete ccmap;
    delete cursor;
}

/** Play the mission. */
void Game::play()
{
    VQAMovie *mov;
    /* start the song */
    SDL_LockAudio();
    sndengine->playSong(ccmap->getMissionData().theme);
    SDL_UnlockAudio();
    /* send a pointer to the map, sidebar and cursor to the gfxengine */
    gfxengine->setupCurrentGame(ccmap, sidebar, cursor);
    ccmap->restoreLocation(0);
    /* main gameloop */
    //   gfxengine->postMessage("Starting main gameloop..");
    while( !input->shouldQuit() ) {
        /* draw the scene */
        gfxengine->renderScene();
        /* run all scheduled events */
        evQueue->runEvents();
        /* handle the input */
        input->handle();
        if (gamemode == 2) {
            /* synchronise events with server */
        }
    }
    /* stop the song */
    sndengine->stopSong();

    if (gamemode == 0) {
        if( ccmap->getPlayerPool()->hasWon() ) {
            try {
                mov = new VQAMovie(gfxengine, sndengine, ccmap->getMissionData().winmov);
                mov->play();
                delete mov;
            } catch (VQAError *vqaerr) {
                delete vqaerr;
            }
        } else if( ccmap->getPlayerPool()->hasLost() ) {
            try {
                mov = new VQAMovie(gfxengine, sndengine, ccmap->getMissionData().losemov);
                mov->play();
                delete mov;
            } catch (VQAError *vqaerr) {
                delete vqaerr;
            }
        }
    }
    dumpstats();
}

void Game::dumpstats()
{
    Player* pl;
    Uint8 h,m,s,i;
    Uint32 uptime = evQueue->getElapsedTime();
    uptime /= 1000;
    h = uptime/3600;
    uptime %= 3600;
    m = uptime/60;
    s = uptime%60;
    logger->renderGameMsg(false);
    logger->gameMsg("Time wasted: %i hour%s%i minute%s%i second%s",
                    h,(h!=1?"s ":" "),m,(m!=1?"s ":" "),s,(s!=1?"s ":" "));
    for (i = 0; i < ccmap->getPlayerPool()->getNumPlayers(); i++) {
        pl = ccmap->getPlayerPool()->getPlayer(i);
        logger->gameMsg("%s\nUnit kills:  %i\n     losses: %i\n"
                        "Structure kills:  %i\n          losses: %i\n",
                        pl->getName(),pl->getUnitKills(),pl->getUnitLosses(),
                        pl->getStructureKills(),pl->getStructureLosses());
    }
}
