// mode: -*- C++ -*-
/*****************************************************************************
 * map.h - header file for map class
 *
 ****************************************************************************/

#ifndef __MAP_H
#define __MAP_H


#ifdef _MSC_VER
#pragma warning (disable: 4503)
#pragma warning (disable: 4786)
#endif


#include <vector>
#include <map>

class CnCMap;
class Unit;

#include "inifile.h"
#include "actioneventqueue.h"
#include "shpimage.h"
#include "unitandstructurepool.h"
#include "playerpool.h"
#include "timeline.h"
#include "shpimage.h"

#define HAS_OVERLAY 0x100
#define HAS_TERRAIN 0x200

/// number of map bookmarks
#define NUMMARKS 5

class LoadingScreen;
class NetConnection;

struct TerrainEntry
{
    Uint32 shpnum;
    Sint16 xoffset;
    Sint16 yoffset;
};

struct TileList
{
    Uint16 templateNum;
    Uint8 tileNum;
};

struct ScrollVector
{
    Sint8 x, y;
    Uint8 t;
};

struct ScrollBookmark
{
    Uint16 x,y;
    Uint16 xtile,ytile;
};

struct ScrollData
{
    Uint16 maxx, maxy;
    Uint16 curx, cury;
    Uint16 maxxtileoffs, maxytileoffs;
    Uint16 curxtileoffs, curytileoffs;
    Uint16 tilewidth;
};

struct MiniMapClipping
{
    Uint16 x,y,w,h;
    Uint16 sidew, sideh;
    Uint8 tilew, tileh, pixsize;
};

struct MissionData
{
    MissionData() {
        theater = brief = action = player = theme = winmov = losemov = mapname = NULL;
    }

    ~MissionData()
    {
        delete[] theater;
        delete[] brief;
        delete[] action;
        delete[] player;
        delete[] theme;
        delete[] winmov;
        delete[] losemov;
        delete[] mapname;
    }
    /// the theater of the map (eg. winter)
    char* theater;
    /// the briefing movie
    char* brief;
    /// the actionmovie
    char* action;
    /// the players side (goodguy, badguy etc.)
    char* player;
    /// specific music to play for this mission
    char* theme;
    /// movie played after completed mission
    char* winmov;
    /// movie played after failed mission
    char* losemov;

    /// the name of the map
    char* mapname;

    /// used to determine what units/structures can be built
    Uint8 buildlevel;
};

class CnCMap
{
public:
    CnCMap(ActionEventQueue *evQueue, NetConnection *connection, LoadingScreen *lscreen);
    ~CnCMap();

    SDL_Surface *getMapTile( Uint32 pos )
    {
        return tileimages[tilematrix[pos]];
    }
    SDL_Surface *getShadowTile(Uint8 shadownum)
    {
        if( shadownum >= numShadowImg ) {
            return NULL;
        }
        return shadowimages[shadownum];
    }
    Uint32 getTiberium(Uint32 pos) const
    {
        return (overlaymatrix[pos] & 0xF);
    }
    Uint32 getSmudge(Uint32 pos) const
    {
        return ((overlaymatrix[pos] & 0xF0) << 12);
    }
    Uint32 setSmudge(Uint32 pos, Uint8 value)
    {
        /// clear the existing smudge bits first
        overlaymatrix[pos] &= ~(0xF0);
        return (overlaymatrix[pos] |= (value<<4));
    }
    Uint32 setTiberium(Uint32 pos, Uint8 value)
    {
        /// clear the existing tiberium bits first
        overlaymatrix[pos] &= ~0xF;
        return (overlaymatrix[pos] |= value);
    }
    Uint32 getOverlay(Uint32 pos);
    Uint32 getTerrain(Uint32 pos, Sint16* xoff, Sint16* yoff);
    Uint8 getTerrainType(Uint32 pos) const
    {
        return terraintypes[pos];
    }

    std::vector<SHPImage *> *getImagePool();

    bool accScroll(Uint8 direction);
    bool absScroll(Sint16 dx, Sint16 dy);
    void doscroll();
    void setMaxScroll(Uint32 x, Uint32 y, Uint32 xtile, Uint32 ytile, Uint32 tilew);
    void setValidScroll();

    const MissionData& getMissionData() const
    {
        return missionData;
    }

    UnitAndStructurePool *getUSPool()
    {
        return uspool;
    }
    Uint32 getScrollPos() const
    {
        return scrollpos.cury*width+scrollpos.curx;
    }
    Uint16 getXScroll() const
    {
        return scrollpos.curx;
    }
    Uint16 getYScroll() const
    {
        return scrollpos.cury;
    }
    Uint16 getXTileScroll() const
    {
        return (Uint16)scrollpos.curxtileoffs;
    }
    Uint16 getYTileScroll() const
    {
        return (Uint16)scrollpos.curytileoffs;
    }
    Uint16 getWidth() const
    {
        return width;
    }
    Uint16 getX() const
    {
        return x;
    }
    Uint16 getY() const
    {
        return y;
    }
    Uint16 getHeight() const
    {
        return height;
    }

    Uint16 getCost(Uint16 pos) const;
    Uint16 getCost(Uint16 pos,Unit* excpUn) const;
    Uint8 getGameMode() const
    {
        return gamemode;
    }
    PlayerPool *getPlayerPool()
    {
        return ppool;
    }
    NetConnection *getNetConnection()
    {
        return connection;
    }
    SDL_Surface* getMiniMap(Uint8 pixside);
    void prepMiniClip(Uint16 sidew, Uint16 sideh)
    {
        miniclip.sidew = sidew;
        miniclip.sideh = sideh;
    }
    const MiniMapClipping& getMiniMapClipping() const {return miniclip;}

    bool toScroll() const
    {
        return toscroll;
    }
    void storeLocation(Uint8 loc);
    void restoreLocation(Uint8 loc);
    TimeLine *getTimeLine()
    {
        return timeline;
    }
    std::vector<Uint16> getWaypoints()
    {
        return waypoints;
    }
    SHPImage* getPips()
    {
        return pips;
    }
    Uint32 getPipsNum() const
    {
        return pipsnum;
    }
    SHPImage* getMoveFlash()
    {
        return moveflash;
    }
    Uint32 getMoveFlashNum() const
    {
        return flashnum;
    }
    Uint32 translateToPos(Uint16 x, Uint16 y) const;
    void translateFromPos(Uint32 pos, Uint16 *x, Uint16 *y) const;
private:
    MissionData missionData;
    /* private functions used to load the map, these are in loadmap.cpp */
    /// load the ini part of the map
    void loadIni();
    /// the map section of the ini
    void simpleSections(INIFile *inifile);
    /// the advanced section of the ini
    void advancedSections(INIFile *inifile);
    /// load the bin part of the map (TD)
    void loadBin();
    /// extract RA map data
    void unMapPack( INIFile* inifile);
    /// load the palette
    void loadPal( SDL_Color *palette );
    /// read the bin part of the map (RA or TD)
    void parseBin(TileList *bindata);
    /// load a specified tile
    SDL_Surface *loadTile( INIFile *templini,
                           Uint16 templ, Uint8 tile, Uint32 &tiletype );

    /// width of map in tiles
    Uint16 width;
    /// height of map in tiles
    Uint16 height;
    /// x coordinate for the first tile
    Uint16 x;
    /// y coordinate for the first tile
    Uint16 y;

    ScrollData scrollpos;
    /* A array of tiles and a vector constining the images for the tiles */
    /// a matrix of the tiles
    Uint16 *tilematrix;
    std::vector<SDL_Surface *>tileimages;
    Uint16 numShadowImg;
    SDL_Surface **shadowimages;

    std::vector<Uint16> waypoints;

    Uint16 *overlaymatrix;
    Uint8 *terraintypes;

    std::vector<SHPImage *> imagepool;
    std::map<Uint32, TerrainEntry> terrains;
    std::map<Uint32, Uint16> overlays;

    Uint8 maptype; // 0: TD 1: RA
    Uint8 gamemode; // 0 - single player, 1 - skirmish, 2 - multiplayer

    SDL_Surface *minimap, *oldmmap;
    MiniMapClipping miniclip;

    ActionEventQueue *evQueue;

    ScrollVector scrollvec;
    bool toscroll;
    /// stores which directions can be scrolled
    Uint8 valscroll;
    /// stores certain map locations
    ScrollBookmark scrollbookmarks[NUMMARKS];
    Uint8 scrollstep, maxscroll, scrolltime;

    UnitAndStructurePool *uspool;
    PlayerPool* ppool;
    NetConnection *connection;
    TimeLine* timeline;

    Uint32 pipsnum;
    SHPImage* pips;
    Uint32 flashnum;
    SHPImage* moveflash;
};

class LoadMapError
    {}
;

#define SCROLL_NONE 0
#define SCROLL_UP 0x1
#define SCROLL_RIGHT 0x2
#define SCROLL_DOWN 0x4
#define SCROLL_LEFT 0x8
#define SCROLL_UPRIGHT (SCROLL_UP | SCROLL_RIGHT)
#define SCROLL_DOWNRIGHT (SCROLL_DOWN | SCROLL_RIGHT)
#define SCROLL_DOWNLEFT (SCROLL_DOWN | SCROLL_LEFT)
#define SCROLL_UPLEFT (SCROLL_UP | SCROLL_LEFT)

#define TTYPE_LAND 0
#define TTYPE_WATER 1
#define TTYPE_ROAD 2
#define TTYPE_ROCK 3
#define TTYPE_TREE 4
#define TTYPE_WATER_BLOCKED 5
#define TTYPE_OTHER_NONPASS 7


#endif

