/*****************************************************************************
 * sidebar.cpp - sidebar class
 *
 ****************************************************************************/
#include <cstdio>
#include <cstdlib>
#include "vfs.h"
#include "sidebar.h"
#include "shpimage.h"
#include "unitandstructurepool.h"
#include "playerpool.h"
#include "input.h"
#include "logger.h"
#include "memtrack.h"

static const char* radarnames[] =
    {
        "hradar.nod", "hradar.gdi", "hradar.jp",
        "radar.nod",  "radar.gdi",  "radar.jp"
    };

/** Constructor, creates the sidebar.
 * @param the mixfiles.
 */
Sidebar::Sidebar( ActionEventQueue* evQueue, Player *pl,
                  Uint16 height, const char *theater)
{
    const char* tmpname;
    Uint8 side;
    SHPImage *tmpshp;
    SDL_Surface *shadow;

    scaleq = -1;

    this->theater = theater;
    player = pl;

    gamefnt = new Font("scorefnt.fnt");
    tmpname = VFS_getFirstExisting(2,"htabs.shp","tabs.shp");
    if (tmpname == NULL) {
        logger->error("Unable to find the tab image! (Maybe you run c&c gold but have forgoten updatec.mix?)\n");
        throw new SidebarError();
    } else if (strcasecmp(tmpname,"htabs.shp")==0) {
        isoriginaltype = false;
    } else {
        isoriginaltype = true;
    }
    try {
        tmpshp = new SHPImage(tmpname, scaleq);
    } catch(ImageNotFound *e) {
        delete e;
        logger->error("Unable to load the tab image!)\n");
        throw new SidebarError();
    }


    tmpshp->getImage(0, &tab, &shadow, 0);
    SDL_FreeSurface(shadow);
    delete tmpshp;

    tablocation.x = 0;
    tablocation.y = 0;
    tablocation.w = tab->w;
    tablocation.h = tab->h;

    spalnum = pl->getStructpalNum();
    if (!isoriginaltype) {
        spalnum = 0;
    }
    if ((player->getSide()&~PS_MULTI) == PS_BAD) {
        side = isoriginaltype?3:0;
    } else if ((player->getSide()&~PS_MULTI) == PS_GOOD) {
        side = isoriginaltype?4:1;
    } else {
        side = isoriginaltype?5:2;
    }
    radarname = radarnames[side];
    try {
        tmpshp = new SHPImage(radarname, scaleq);
    } catch(ImageNotFound *e) {
        delete e;
        logger->error("Hmm.. managed to misdetect sidebar type\n");
        try {
            if (side < 3)
                side += 3;
            else
                side -= 3;
            radarname = radarnames[side];
            tmpshp = new SHPImage(radarname, scaleq);
        } catch(ImageNotFound *e) {
            delete e;
            logger->error("Unable to load the radar-image! (Maybe you run c&c gold but have forgoten updatec.mix?)\n");
            throw new SidebarError();
        }
    }

    tmpshp->getImage(0, &radarlogo, &shadow, 0);
    SDL_FreeSurface(shadow);
    delete tmpshp;

    visible = 1;
    changed = 1;

    unitoff = 0;
    structoff = 0;

    buttondown = 0;
    bd = false;

    sbar = NULL;
    sbarlocation.x = sbarlocation.y = sbarlocation.w = sbarlocation.h = 0;

    this->evQueue = evQueue;

    this->radaranim = NULL;
    radaranimating = false;

    setupbuttons(height);
}

/** Destructor, free all memory used by the sidebar.*/
Sidebar::~Sidebar()
{
    Uint32 i;
    SDL_FreeSurface(tab);
    SDL_FreeSurface(radarlogo);
    if( sbar != NULL )
        SDL_FreeSurface(sbar);
    delete gamefnt;
    for( i = 0; i < buttons.size(); i++ ) {
        delete buttons[i];
    }
    for (i=0;i<uniticons.size();++i) {
        delete[] uniticons[i];
    }
    uniticons.resize(0);
    for (i=0;i<structicons.size();++i) {
        delete[] structicons[i];
    }
    structicons.resize(0);
}

/** Check if the sidebar has changed and needs to be redrawn.
 * By changed we mean has become visible or hidden.
 * @returns 1 if the sidebar has changed, 0 otherwise.
 */
Uint8 Sidebar::getChanged()
{
    if( changed > 0 ) {
        changed = 0;
        return 1;
    }
    return 0;
}

/** Toggles whether sidebar is visible. */
void Sidebar::toggleVisible()
{
    if( visible == 1 )
        visible = 0;
    else
        visible = 1;
    changed = 1;
}

SDL_Surface *Sidebar::getSidebarImage(SDL_Rect location)
{
    SDL_Rect dest, src;
    SDL_Surface *temp, *shadow;
    SHPImage *textureshp;


    if( location.w == sbarlocation.w && location.h == sbarlocation.h )
        return sbar;

    if( sbar != NULL )
        SDL_FreeSurface(sbar);

    temp = SDL_CreateRGBSurface(SDL_SWSURFACE, location.w, location.h, 16, 0, 0, 0, 0);
    sbar = SDL_DisplayFormat(temp);
    SDL_FreeSurface(temp);
    location.x = 0;
    location.y = 0;

    if (isoriginaltype) {
        SDL_FillRect(sbar, &location, SDL_MapRGB(sbar->format, 0xa0, 0xa0, 0xa0));
    } else {
        try {
            textureshp = new SHPImage("btexture.shp", scaleq);
            textureshp->getImage(1, &temp, &shadow, 0);
            delete textureshp;
            SDL_FreeSurface(shadow);
            dest.x = 0;
            dest.w = location.w;
            dest.h = temp->h;
            src.x = 0;
            src.y = 0;
            src.h = temp->h;
            src.w = location.w;
            for( dest.y = 0; dest.y < location.h; dest.y += dest.h ) {
                SDL_BlitSurface( temp, &src, sbar, &dest );
            }
            SDL_FreeSurface(temp);
        } catch(ImageNotFound *e) {
            delete e;
            logger->error("Hmm.. possible misdetction or corrupt file\n");
            SDL_FillRect(sbar, &location, SDL_MapRGB(sbar->format, 0xa0, 0xa0, 0xa0));
        }
    }
    sbarlocation = location;

    if (!Input::isMinimapEnabled()) {
        dest.x = 0;
        dest.y = 0;
        dest.w = radarlogo->w;
        dest.h = radarlogo->h;
        SDL_BlitSurface( radarlogo, NULL, sbar, &dest);
    }
    for (Sint8 x=buttons.size()-1;x>=0;--x) {
        dest = buttons[x]->getRect();
        temp = buttons[x]->getSurface();
        if (sbar != NULL)
            SDL_FillRect(sbar,&dest,0x0);
        //SDL_SetClipRect(temp,&dest);
        SDL_BlitSurface(temp, NULL, sbar, &dest);
    }

    return sbar;
}

void Sidebar::addButton(Uint16 x, Uint16 y, const char* fname, Uint8 f, Uint8 pal)
{
    SidebarButton *tmp = new SidebarButton(x,y,fname,f,theater,this, pal);
    buttons.push_back(tmp);
    changed = 1;
}

void Sidebar::setupbuttons(Uint16 height)
{
    const char* tmpname;
    Uint16 scrollbase;
    Uint8 t;

    Uint32 striph,stripw;
    Uint32 startoffs = tab->h+radarlogo->h;

    SHPImage *strip;

    tmpname = VFS_getFirstExisting(2,"hstrip.shp","strip.shp");
    if (tmpname == NULL) {
        logger->error("Unable to find strip images for sidebar, exiting\n");
        throw new SidebarError();
    }
    try {
        strip = new SHPImage(tmpname, scaleq);
    } catch(ImageNotFound *e) {
        delete e;
        logger->error("Unable to load strip images for sidebar, exiting\n");
        throw new SidebarError();
    }

    striph = strip->getHeight();
    stripw = strip->getWidth();

    delete strip;

    if( striph > 100 )
        striph = striph>>2;

    buildbut = ((height-startoffs)/striph)-2;
    startoffs += striph;

    scrollbase = startoffs + striph*buildbut;
    addButton(10+stripw,scrollbase,"stripup.shp",B_SCROLL|B_UNIT|B_UP,0); // 0
    addButton(10,scrollbase,"stripup.shp",B_SCROLL|B_STRUCTURE|B_UP,0); // 1

    addButton(10+stripw+(stripw>>1),scrollbase,"stripdn.shp",B_SCROLL|B_UNIT|B_DOWN,0); // 2
    addButton(10+(stripw>>1),scrollbase,"stripdn.shp",B_SCROLL|B_STRUCTURE|B_DOWN,0); // 3

    // The order in which the addButton calls are made MUST be preserved
    // Two loops are made so that all unit buttons and all structure buttons
    // are grouped contiguously (4,5,6,7,...) compared to (4,6,8,10,...)

    for (t=0;t<buildbut;++t) {
        addButton(10+stripw,startoffs+striph*t,"strip.shp",B_BUILD|B_UNIT,0);
    }
    for (t=0;t<buildbut;++t) {
        addButton(10,startoffs+striph*t,"strip.shp",B_BUILD|B_STRUCTURE,spalnum);
    }

    //createdummylists();
    updateAvailableLists();

    updateicons();
    if (uniticons.size() == 0 && structicons.size() == 0) {
        visible = 0;
        changed = 1;
    }
}

Uint8 Sidebar::getButton(Uint16 x,Uint16 y)
{
    SDL_Rect tmp;

    for (Uint8 i=0;i<buttons.size();++i) {
        tmp = buttons[i]->getRect();
        if (x>=tmp.x && y>=tmp.y && x<(tmp.x+tmp.w) && y<(tmp.y+tmp.h)) {
            return i;
        }
    }
    return 255;
}

/** Event handler for clicking buttons.
 * 
 * @bug This is an evil hack that should be replaced by something more flexible.
 */
void Sidebar::clickButton(Uint8 index, char* unitname, Uint8* createmode)
{
    Uint8 f = buttons[index]->getFunction();

    switch (f&0x3) {
    case 0:
        return;
        break;
    case 1: // build
        build(index - 3 - ((f&B_UNIT)?0:buildbut),(f&B_UNIT),unitname, createmode);
        break;
    case 2: // scroll
        downbutton(index);
        scrollbuildlist( (f&B_UP), (f&B_UNIT) );
        break;
    default:
        logger->error("Sidebar::clickButton. This should not happen (%i)\n",f&0x3);
        break;
    }
}

void Sidebar::build(Uint8 index, Uint8 type, char* unitname, Uint8* createmode)
{
    Uint8* offptr;
    std::vector<char*>* vecptr;

    if (type) {
        offptr = &unitoff;
        vecptr = &uniticons;
        *createmode = 1;
    } else {
        offptr = &structoff;
        vecptr = &structicons;
        *createmode = 0;
    }

    if ( (unsigned)(*vecptr).size() > ((unsigned)(*offptr)+index - 1) ) {
        strncpy(unitname,(*vecptr)[(*offptr+index-1)],13);
        unitname[strlen((*vecptr)[(*offptr+index-1)])-8] = 0x0;
    }
}

void Sidebar::scrollbuildlist(Uint8 dir, Uint8 type)
{
    Uint8* offptr;
    std::vector<char*>* vecptr;

    if (type) {
        offptr = &unitoff;
        vecptr = &uniticons;
    } else {
        offptr = &structoff;
        vecptr = &structicons;
    }

    if (dir) { //up
        if (*offptr>0) {
            --(*offptr);
        }
    } else { // down
        if ((unsigned)(*offptr+buildbut) < (unsigned)vecptr->size()) {
            ++(*offptr);
        }
    }

    updateicons();
}

void Sidebar::updateSidebar()
{
    updateAvailableLists();
    updateicons();
}

/** Rebuild the list of icons that are available.
 * 
 * @bug Newer items should be appended, although with some grouping (i.e. keep
 * infantry at top, vehicles, aircraft, boats, then superweapons).
 */
void Sidebar::updateAvailableLists(void)
{
    std::vector<const char*> units_avail, structs_avail;
    char* nametemp;
    bool played_sound;
    Uint32 i;
    played_sound = false;
    units_avail = player->getUSPool()->getBuildableUnits(player);
    structs_avail = player->getUSPool()->getBuildableStructures(player);
    if (units_avail.size() != uniticons.size()) {
        unitoff = 0;
        if (units_avail.size() > uniticons.size()) {
            played_sound = true;
            player->getUSPool()->getsfxeng()->queueSound("newopt1.aud");
        }
        for (i=0;i<uniticons.size();++i) {
            delete[] uniticons[i];
        }
        uniticons.resize(0);
        for (i=0;i<units_avail.size();++i) {
            nametemp = new char[13];
            memset(nametemp,0x0,13);
            sprintf(nametemp,"%sICON.SHP",units_avail[i]);
            uniticons.push_back(nametemp);
        }
    }
    if (structs_avail.size() != structicons.size()) {
        structoff = 0;
        if ((structs_avail.size() > structicons.size())&&!played_sound) {
            player->getUSPool()->getsfxeng()->queueSound("newopt1.aud");
        }
        for (i=0;i<structicons.size();++i) {
            delete[] structicons[i];
        }
        structicons.resize(0);
        for (i=0;i<structs_avail.size();++i) {
            nametemp = new char[13];
            memset(nametemp,0x0,13);
            sprintf(nametemp,"%sICON.SHP",structs_avail[i]);
            structicons.push_back(nametemp);
        }
    }
    if (uniticons.size() == 0 && structicons.size() == 0) {
        visible = 0;
        changed = 1;
    } else {
        visible = 1;
        changed = 1;
    }
}

/** Sets the images of the visible icons, having scrolled.
 */
void Sidebar::updateicons(void)
{
    Uint8 i;
    SDL_Rect dest;

    for (i=0;i<buildbut;++i) {
        if ((unsigned)(i+unitoff)>=(unsigned)uniticons.size()) {
            buttons[4+i]->ChangeImage("strip.shp");
        } else {
            if (!buttons[4+i]->ChangeImage(uniticons[i+unitoff])) {
                buttons[4+i]->ChangeImage("strip.shp");
            }
        }
    }
    for (i=0;i<buildbut;++i) {
        if ((unsigned)(i+structoff)>=(unsigned)structicons.size()) {
            buttons[buildbut+4+i]->ChangeImage("strip.shp");
        } else {
            if (!buttons[buildbut+4+i]->ChangeImage(structicons[i+structoff])) {
                buttons[buildbut+4+i]->ChangeImage("strip.shp");
            }
        }
    }

    for (Sint8 x=buttons.size()-1;x>=0;--x) {
        dest = buttons[x]->getRect();
        if (sbar != NULL)
            SDL_FillRect(sbar,&dest,0x0);
        //SDL_SetClipRect(buttons[x]->getSurface(),&dest);
        SDL_BlitSurface(buttons[x]->getSurface(), NULL, sbar, &dest);
    }
}

void Sidebar::downbutton(Uint8 index)
{
    if (index>3)
        return; // not a scroll button

    buttondown = index;

    bd = true;
    if (index == 0 || index == 1)
        buttons[index]->ChangeImage("stripup.shp",1);
    else
        buttons[index]->ChangeImage("stripdn.shp",1);
    updateicons();

}

void Sidebar::resetButton(void)
{
    if (!bd)
        return;
    bd = false;

    if (buttondown < 2)
        buttons[buttondown]->ChangeImage("stripup.shp",0);
    else
        buttons[buttondown]->ChangeImage("stripdn.shp",0);
    buttondown = 0;
    updateicons();
}

void Sidebar::startRadarAnim(Uint8 mode, bool* minienable)
{
    if (radaranimating == false && radaranim == NULL && sbar != NULL) {
        radaranimating = true;
        radaranim = new RadarAnimEvent(2,mode,this,minienable);
    }
}

void Sidebar::scrollSidebar(bool scrollup)
{
    scrollbuildlist(scrollup, 0);
    scrollbuildlist(scrollup, 1);
}

SidebarButton::SidebarButton(Sint16 x, Sint16 y, const char* mixname,
                             Uint8 f, const char* theater, Sidebar* sb,
                             Uint8 pal)
{
    pic = NULL;
    picloc.x = x;
    picloc.y = y;
    function = f;
    palnum = pal;
    this->theater = theater;
    sidebar = sb;
    ChangeImage(mixname);
}

bool SidebarButton::ChangeImage(const char* fname)
{
    return ChangeImage(fname,0);
}

void SidebarButton::fallback(const char* fname)
{
    Uint32 width, height;
    Uint32 slen = strlen(fname);
    char* iname = new char[slen-7];
    strncpy(iname,fname,slen-8);
    iname[slen-8] = 0;
    width = sidebar->getFont()->calcStringWidth(iname);
    height = sidebar->getFont()->getHeight()+1;
    pic = SDL_CreateRGBSurface(SDL_SWSURFACE|SDL_SRCCOLORKEY, width, height, 16, 0, 0, 0, 0);
    SDL_FillRect(pic, NULL, 0);
    sidebar->getFont()->drawString(iname,pic,0,0);
    delete[] iname;
}

bool SidebarButton::ChangeImage(const char* fname, Uint8 number)
{
    SHPImage* tmpshp;
    SDL_Surface* shadow;

    int scaleq = -1;
    char goldname[32];
    Uint32 slen = strlen(fname);

    if (pic != NULL)
        SDL_FreeSurface(pic);

    if( slen>8 && strcasecmp("icon.shp", fname+slen-8)==0 ) {
        if (sidebar->isOriginalType()) {
            try {
                tmpshp = new SHPImage(fname, scaleq);
            } catch (ImageNotFound* e) {
                fallback(fname);
                return true;
            }
        } else {
            strcpy(goldname, fname);
            goldname[slen-6] = 'n';
            goldname[slen-5] = 'h';
            goldname[slen-3] = theater[0];
            goldname[slen-2] = theater[1];
            goldname[slen-1] = theater[2];
            try {
                tmpshp = new SHPImage(goldname, scaleq);
            } catch( ImageNotFound *e) {
                delete e;
                try {
                    tmpshp = new SHPImage(fname, scaleq);
                } catch( ImageNotFound* e2) {
                    delete e2;
                    fallback(fname);
                    return true;
                }
            }
        }
    } else {
        if (sidebar->isOriginalType()) {
            try {
                tmpshp = new SHPImage(fname, scaleq);
            } catch (ImageNotFound *e3) {
                fallback(fname);
                return true;
            }
        } else {
            sprintf(goldname, "h%s", fname);
            try {
                tmpshp = new SHPImage(goldname, scaleq);
            } catch( ImageNotFound *e) {
                fallback(fname);
                return true;
            }
        }
    }

    tmpshp->getImage(number,&pic,&shadow,palnum);
    SDL_FreeSurface(shadow);
    delete tmpshp;

    picloc.w = pic->w;
    picloc.h = pic->h;

    if (picloc.h > 100) {
        picloc.h >>=2;
        pic->h >>= 2;
    }
    return true;
}

SidebarButton::~SidebarButton()
{
    if (pic != NULL) {
        SDL_FreeSurface(pic);
    }
}


RadarAnimEvent::RadarAnimEvent(Uint32 p, Uint8 mode, Sidebar* sidebar, bool* minienable) : ActionEvent(p)
{
    this->mode = mode;
    this->sidebar = sidebar;
    this->minienable = minienable;
    switch (mode) {
    case 0:
        frame = 0;
        framend = 20;
        break;
    case 1:
        //*minienable = false;
        frame = 20;
        framend = 30;
        break;
    default:
        frame = 0;
    }
    sidebar->evQueue->scheduleEvent(this);
}

void RadarAnimEvent::run()
{
    SDL_Rect dest;
    SDL_Surface *shadow;
    SHPImage* tmpshp;

    if (frame <= framend) {
        tmpshp = new SHPImage(sidebar->radarname, sidebar->scaleq);

        dest.x = 0;
        dest.y = 0;
        dest.w = sidebar->radarlogo->w;
        dest.h = sidebar->radarlogo->h;

        SDL_FillRect(sidebar->radarlogo, &dest, SDL_MapRGB(sidebar->sbar->format, 0x0a, 0x0a, 0x0a));
        SDL_BlitSurface( sidebar->radarlogo, NULL, sidebar->sbar, &dest);

        tmpshp->getImage(frame, &(sidebar->radarlogo), &shadow, 0);
        SDL_BlitSurface( sidebar->radarlogo, NULL, sidebar->sbar, &dest);

        SDL_FreeSurface(shadow);
        delete tmpshp;
        ++frame;
        sidebar->evQueue->scheduleEvent(this);
    } else {
        dest.x = 0;
        dest.y = 0;
        dest.w = sidebar->radarlogo->w;
        dest.h = sidebar->radarlogo->h;

        if (mode == 1) {
            tmpshp = new SHPImage(sidebar->radarname, sidebar->scaleq);
            tmpshp->getImage(0, &(sidebar->radarlogo), &shadow, 0);
            SDL_BlitSurface( sidebar->radarlogo, NULL, sidebar->sbar, &dest);
            SDL_FreeSurface(shadow);
            delete tmpshp;
            sidebar->radaranim = NULL;
        } else {
            SDL_FillRect(sidebar->radarlogo, &dest, SDL_MapRGB(sidebar->sbar->format, 0x0a, 0x0a, 0x0a));
            SDL_BlitSurface( sidebar->radarlogo, NULL, sidebar->sbar, &dest);
            *minienable = true;
            sidebar->radaranim = NULL;
        }
        sidebar->radaranimating = false;
        delete this;
        return;
    }
}
